library(data.table)
library(dplyr)

# globals ----------------------------------------------------------------------
years_unplanned <- 10
years_planned <- 7
mes_per_year <- 2.3

small_edb_invercargill <- 734  # 726 for DPP3
small_edb_nelson <- 328        # 329 for DPP3

saidi_collar_rateup <- 2
saifi_collar_rateup <- 2
saidi_cap_rateup <- 5
saifi_cap_rateup <- 5

# utils ------------------------------------------------------------------------
source("R/utils.R")

inputs_l <- list(
  input_file="data-raw/dpp4_edb_interruptions_transitional (5269107.1).xlsx",
  start_=as.POSIXct("2014-03-31 00:00:00", tz = "UTC"),
  end_=as.POSIXct('2024-03-31 23:30:00', tz = "UTC")
)

# ------------------------------------------------------------------------------
input_file_rds_cache <- paste0(
  "outputs/", sub(pattern = "(.*)\\..*$", replacement = "\\1", basename(inputs_l[["input_file"]])), ".rds")

# import data ------------------------------------------------------------------
## DPP4 data doesn't import correctly -- need to manually deal with date

dpp <- readxl::read_xlsx(
  inputs_l[["input_file"]],
  col_types = c(
    "text", "numeric", "text", "text", "text",
    "numeric", "numeric", "numeric", "numeric"
  )
) |>
  rename_cols()

v <- as_date(dpp$start_date_and_time)
stopifnot(sum(is.na(v)) == 0)
dpp$start_date_and_time <- v

dpp <- dpp |>
  rename(
    planned_or_unplanned = planned_unplanned
  )

dpp <- dpp |>
  mutate( # RP: fix discrepancies introduced between draft and after
    planned_or_unplanned = tolower(planned_or_unplanned)
  )

# half hours -------------------------------------------------------------------
all_halfhours <- data.frame(time = seq(inputs_l[["start_"]], inputs_l[["end_"]], 60 * 30))

# boundaries -------------------------------------------------------------------
boundaries <- dpp |> select(edb) |> distinct() |>
  mutate(
    boundary = case_when(
      edb == "Electricity Invercargill" ~ small_edb_invercargill,
      edb == "Nelson Electricity" ~ small_edb_nelson,
      TRUE ~ mes_per_year * 48 * years_unplanned
    )
  )

# unplanned boundary values ----------------------------------------------------
# cut events for dates outside range
# create time variable rounded down to nearest half hour
unplanned <- dpp |>
  filter(planned_or_unplanned == "unplanned") |>
  filter(as.Date(start_date_and_time) >= as.Date(inputs_l[["start_"]])) |>
  filter(as.Date(start_date_and_time) <= as.Date(inputs_l[["end_"]])) |>
  mutate(
    time = round_down_to_nearest_half_hour(start_date_and_time)
  ) |>
  select(edb, time, saidi_value, saifi_value)

# create scaffold--all unique edb x half-hour combinations
edb_halfhours <- unplanned |>
  select(edb) |> distinct() |> cross_join(all_halfhours)

# merge unplanned with scaffold
unplanned_ <- left_join(edb_halfhours, unplanned, by = c("edb", "time")) |>
  arrange(edb, time) |>
  mutate(
    saidi_value = ifelse(is.na(saidi_value), 0, saidi_value),
    saifi_value = ifelse(is.na(saifi_value), 0, saifi_value)
  )

# get sum of saidi and saifi by half-hour
half_hour_totals <- unplanned_ |>
  group_by(edb, time) |>
  summarise(
    saidi = sum(saidi_value),
    saifi = sum(saifi_value),
    .groups = "drop"
  )

# get number of half-hour periods with a non-zero saidi value
half_hour_counts <- half_hour_totals |>
  group_by(edb) |>
  summarise(n = sum(saidi > 0))

# get rolling 24-hour totals
# cumsum(x)[t] - cumsum(x)[t -]
rolling_total <- half_hour_totals |>
  group_by(edb) |>
  mutate(
    saidi_roll = cumsum(saidi) - lag(cumsum(saidi), 48),
    saifi_roll = cumsum(saifi) - lag(cumsum(saifi), 48)
  ) |>
  ungroup()

# get saidi boundary values
boundary_saidi <-
  get_boundary(rolling_total, "edb", "saidi_roll", "saidi_bound", boundaries)

# get saifi boundary values
boundary_saifi <-
  get_boundary(rolling_total, "edb", "saifi_roll", "saifi_bound", boundaries)

# merge results
boundary_values <- half_hour_counts |>
  left_join(boundary_saidi, by = "edb") |>
  left_join(boundary_saifi, by = "edb")

## normalise -------------------------------------------------------------------
half_hour_cumsum <- rolling_total |>
  mutate(
    date = as.Date(time),
    year = year(time),
    month = month(time)
  ) |>
  mutate(
    disc_year = ifelse(month >= 4, year + 1, year)
  ) |>
  group_by(edb) |>
  mutate(
    cum_saidi = cumsum(saidi),
    cum_saifi = cumsum(saifi)
  ) |>
  left_join(
    boundary_values, by = "edb"
  )

# Part of the code published for the draft decision has been replaced below
# to use data.table rather than long inefficient loops.
# It's easier to validate and faster to run!
hfc_dt <- as.data.table(half_hour_cumsum)  # convert to data.table

# initialise variables for normalisation
hfc_dt[, `:=`(saidi_norm = saidi, saifi_norm = saifi, saidi_major = 0, saifi_major = 0)]

# perform normalisation: 1) rolling assessment  2) lower values if major event
hfc_dt[, max_saidi := frollapply(
  x = saidi_roll, n = 48, max, fill = NA, align = "left", na.rm = TRUE), by=edb]
hfc_dt[max_saidi > saidi_bound, `:=`(
  saidi_norm=pmin(saidi, saidi_bound/48), saidi_major=1)]

hfc_dt[, max_saifi := frollapply(
  x = saifi_roll, n = 48, max, fill = NA, align = "left", na.rm = TRUE), by=edb]
hfc_dt[max_saifi > saifi_bound, `:=`(
  saifi_norm=pmin(saifi, saifi_bound/48), saifi_major=1)]

# only keep data in chosen regulatory years (more initially kept for moving observation windows)
hfc_dt <- hfc_dt[disc_year %in% sort(unique(hfc_dt$disc_year), decreasing=T)[1:years_unplanned]]

# compute target and cap
target_dt <- hfc_dt[, .(saidi_target = sum(saidi_norm)/years_unplanned,
                        saifi_target = sum(saifi_norm)/years_unplanned), by=edb]
sd_dt <- hfc_dt[, .(saidi_daily_norm = sum(saidi_norm),
                    saifi_daily_norm = sum(saifi_norm)), by=c("edb", "date")][
                      , .(saidi_sd = sd(saidi_daily_norm) * sqrt(365.25),
                          saifi_sd = sd(saifi_daily_norm) * sqrt(365.25)), by=c("edb")]
targets <- merge(target_dt, sd_dt)[, `:=`(saidi_cap = saidi_target + 2 * saidi_sd,
                                         saifi_cap = saifi_target + 2 * saifi_sd)]

# sanity check, did we lose data? edbs have interruptions every year right?
annual_dt <- hfc_dt[, .(assessed_saidi = sum(saidi_norm), assessed_saifi = sum(saifi_norm)), by=c("edb", "disc_year")]
stopifnot(
  nrow(
    annual_dt[(assessed_saidi == 0 | assessed_saifi == 0) &
                ! (edb == "Nelson Electricity" & disc_year == 2021)]  # list here if okay not to have interruptions
  ) == 0)


## final unplanned parameters --------------------------------------------------
unplanned_parameters <- boundary_values |>
  inner_join(targets, by = "edb") |>
  mutate(saidi_collar = 0, saifi_collar = 0) |>
  select(
    edb, n,
    saidi_bound, saidi_target, saidi_collar, saidi_sd, saidi_cap,
    saifi_bound, saifi_target, saifi_collar, saifi_sd, saifi_cap
  )

write.csv(
  unplanned_parameters,
  paste0("outputs/unplanned_norm_param_not_caps_adjusted", "[",
         basename(inputs_l[["input_file"]]), "]", ".csv"),
  row.names = FALSE
)

# planned parameters -----------------------------------------------------------
planned <- dpp |>
  filter(planned_or_unplanned == "planned") |>
  filter(disclosure_year >= year(inputs_l[["start_"]])) |>
  filter(start_date_and_time >= round_up_to_nearest_month(inputs_l[["start_"]])) |>
  filter(start_date_and_time <= inputs_l[["end_"]])

years <- planned$disclosure_year |> unique() |> sort() |> tail(years_planned)

planned_parameters <- planned |>
  filter(disclosure_year %in% years) |>
  group_by(edb) |>
  summarise(
    saidi_target = sum(saidi_value, na.rm = TRUE) / years_planned,
    saifi_target = sum(saifi_value, na.rm = TRUE) / years_planned,
    .groups = "drop"
  ) |>
  mutate(
    saidi_collar = 0,
    saifi_collar = 0,
    saidi_cap = saidi_collar_rateup * saidi_target,
    saifi_cap = saifi_collar_rateup * saifi_target,
    saidi_limit = saidi_cap_rateup * saidi_cap,
    saifi_limit = saifi_cap_rateup * saifi_cap
  ) |>
  select(
    edb,
    saidi_collar, saidi_target, saidi_cap, saidi_limit,
    saifi_collar, saifi_target, saifi_cap, saifi_limit
  )

write.csv(
  planned_parameters,
  paste0("outputs/planned_param_not_caps_adjusted", "[",
         basename(inputs_l[["input_file"]]), "]", ".csv"),
  row.names = FALSE
)
