library(dplyr)

#' Get year from a date
#'
#' Get year from a date.  Removes the need to include the lubridate package.
#'
#' @param x a date, POSIXct, etc.
#' @returns year as integer
year <- function(x) {
  as.integer(format(x, "%Y"))
}

#' Get month from a date
#'
#' Get month from a date.  Removes the need to include the lubridate package.
#'
#' @param x a date, POSIXct, etc.
#' @returns year as integer
month <- function(x) {
  as.integer(format(x, "%m"))
}

#' Round a datetime up to nearest month.
#'
#' @param x a datetime variable
#' @returns a datetime variable rounded up to nearest month
round_up_to_nearest_month <- function(x, tz = "utc") {
  y <- year(x)
  m <- month(x) + 1
  d <- 1
  as.POSIXct(sprintf("%04d-%02d-%02d 00:00:000", y, m, d), tz = tz)
}

#' Round a datetime down to nearest whole half hour.
#'
#' @param x a datetime variable.
#' @returns a datetime variable, rounded to nearest whole whole half hour.
round_down_to_nearest_half_hour <- function(x) {
  sapply(x, \(y) {
    floor(as.numeric(y) / 1800) * 1800
  }) |> as.POSIXct(tz = "utc")
}

#' Rename dataframe columns
#'
#' Rename columns by replacing whitespace with underscores, and making lower case.
#'
#' @param x a dataframe
#' @returns a dataframe with columns renamed
rename_cols <- function(x) {
  setNames(x, gsub("\\s+", "_", tolower(colnames(x))))
}

#' Get records corresponding to some boundary value
#'
#' @param byvar string containing name of grouping variable, defaults to `edb`
#' @param invar string containing name of sorting variable
#' @param outvar string containing name of output variable
#' @param boundaries a dataframe containing index of boundary value.  Must contain `byvar`.
get_boundary <- function(data, byvar = "edb", invar, outvar, boundaries) {
  x <- data |>
    arrange(!!rlang::sym(byvar), desc(!!rlang::sym(invar)), time) |>
    group_by(!!rlang::sym(byvar)) |>
    mutate(
      rank_ = row_number()
    ) |>
    inner_join(
      boundaries, by = byvar
    ) |>
    filter(
      rank_ == round(boundary, 0)
    ) |>
    select(
      all_of(byvar), all_of(invar)
    ) |>
    rename(
      !! outvar := invar
    )
}


#' Convert vector sourced from Excel to date
#'
#' @param x vector of strings
#' @returns a vector of dates
as_date <- function(x) {

  # RP: simplified reimplementation
  # allows for input to be an excel datetime or a string with "%d/%m/%Y %H:%M" format
  dt_tmp <- data.table()
  suppressWarnings({
    dt_tmp[, date := openxlsx::convertToDateTime(x)]
    dt_tmp[, date2 := as.POSIXct(x, format="%d/%m/%Y %H:%M", tz="UTC")]
  })
  dt_tmp[is.na(date), date := date2]
  if (sum(is.na(dt_tmp$date))) warning(
    "Some NA persist after trying our very best to convert to a datetime")
  return(dt_tmp$date)

}
